@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2008, 2010-2012, 2014-2017 Free Software Foundation,
-   Inc.
+   Copyright (C) 2008-2020 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -33,3 +32,12 @@ int manlock_unlock (mu_mailbox_t mbox);
 
 void manlock_cfg_init (void);
 
+void unistr_downcase (char const *input, char **output);
+int unistr_is_substring (char const *haystack, char const *needle);
+int unistr_is_substring_dn (char const *haystack, char const *needle);
+
+int message_body_stream (mu_message_t msg, int unix_header,
+			 char const *charset,
+			 mu_stream_t *pstr);
+extern struct mu_cli_capa mu_cli_capa_mime;
+
