@@ -1099,7 +1099,7 @@ append_message_to_stream (mu_stream_t ostr, mu_message_t msg,
   mu_stream_unref (istr);
   if (status == 0)
     {
-      status = mu_stream_copy (ostr, flt, 0, NULL);
+      status = mu_stream_copy_nl (ostr, flt, 0, NULL);
       mu_stream_destroy (&flt);
       if (status == 0)
 	status = mu_stream_write (ostr, "\n", 1, NULL);
@@ -1270,6 +1270,9 @@ mbox_expunge_unlocked (mu_mailbox_t mailbox, size_t dirty, int remove_deleted,
 	    }
 	  status = mu_stream_copy (tempstr, mailbox->stream,
 				   mum->body_end - mum->envel_from, NULL);
+	  /* Terminate message with an empty line */
+	  if (status == 0)
+	    status = mu_stream_write (tempstr, "\n", 1, NULL);      
 	  if (status)
 	    {
 	      mu_error (_("%s:%d: error copying: %s"),
